#include <Alert.h>
#include <Entry.h>
#include <Path.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h> // debug
#include "main.h"
#include "defs.h"
#include "beosmain.h"

#define kMemPoolSize	16
#define	kBlockSize	sizeof(W_Event)


GalagaView::GalagaView(BRect frame):
    BView(frame, "view", B_FOLLOW_ALL, B_WILL_DRAW),
    _eventLock(), _eventList(), _memPool(kMemPoolSize, kBlockSize, B_OBJECT_CACHE)
{
    SetViewColor(0, 0, 0);
}


bool
GalagaView::eventsPending()
{
    bool pending;

    _eventLock.Lock();
    pending = !_eventList.IsEmpty();
    _eventLock.Unlock();

    return pending;
}


void
GalagaView::nextEvent(W_Event *event)
{
    _eventLock.Lock();

    if (!_eventList.IsEmpty()) {
	void *p = (void *)_eventList.FirstItem();
	memcpy(event, p, sizeof(W_Event));
	_eventList.RemoveItem(p);
	_memPool.Save(p, kBlockSize);
    }

    _eventLock.Unlock();
}


void
GalagaView::Draw(BRect r)
{
}


void
GalagaView::KeyDown(const char *bytes, int32 numBytes)
{
    if (numBytes == 1) {
	W_Event *event = (W_Event *)_memPool.Get(kBlockSize);
	event->type = W_EV_KEY;
	event->x = event->y = 0;
	switch (bytes[0]) {
	    case B_LEFT_ARROW:
		event->key = 128 + 'b';
		break;
	    case B_RIGHT_ARROW:
		event->key = 128 + 'f';
		break;
	    case B_SPACE:
		event->key = ' ';
		break;
	    default:
		event->key = bytes[0];
		break;
	}

	_eventLock.Lock();
	_eventList.AddItem((void*)event);
	_eventLock.Unlock();
    }
}


void
GalagaView::KeyUp(const char *bytes, int32 numBytes)
{
    if (numBytes == 1) {
	W_Event *event = (W_Event *)_memPool.Get(kBlockSize);
	event->type = W_EV_KEY_OFF;
	event->x = event->y = 0;
	switch (bytes[0]) {
	    case B_LEFT_ARROW:
		event->key = 128 + 'b';
		break;
	    case B_RIGHT_ARROW:
		event->key = 128 + 'f';
		break;
	    case B_SPACE:
		event->key = ' ';
		break;
	    default:
		event->key = bytes[0];
		break;
	}

	_eventLock.Lock();
	_eventList.AddItem((void*)event);
	_eventLock.Unlock();
    }
}


void
GalagaView::MouseDown(BPoint p)
{
    W_Event *event = (W_Event *)_memPool.Get(kBlockSize);
    event->type = W_EV_BUTTON;
    event->x = (int)p.x;
    event->y = (int)p.y;
    event->key = 0;

    _eventLock.Lock();
    _eventList.AddItem((void*)event);
    _eventLock.Unlock();
}


GalagaWin::GalagaWin():
    BWindow(BRect(50,50,60,60), "XGalaga", B_TITLED_WINDOW,
	B_NOT_RESIZABLE)
{
    extern W_Window baseWin;

    ResizeTo(WINWIDTH, WINHEIGHT);
    GalagaView *view = new GalagaView(Bounds());
    AddChild(view);
    view->MakeFocus();
    Show();

    BBitmap *bitmap = new BBitmap(Bounds(), B_CMAP8, true);
    memset(bitmap->Bits(), 0, bitmap->BitsLength());
    BView *buffer = new BView(Bounds(), "buffer", B_FOLLOW_ALL, B_WILL_DRAW);
    buffer->SetLowColor(0, 0, 0);
    bitmap->AddChild(buffer);

    baseWin = new struct _W_Window;
    baseWin->view = (char *)view;
    baseWin->buffer = (char *)buffer;
    baseWin->bitmap = (char *)bitmap;
    _zoomed = false;
}


GalagaWin::~GalagaWin()
{
    extern W_Window baseWin;

    BBitmap *b = (BBitmap *)baseWin->bitmap;
    delete b;
}


bool
GalagaWin::QuitRequested()
{
    extern int quitting;
    static bool called = false;

    if (quitting) {
	called = true;
    }
    else {
	quitting = 1;
	snooze(100000); // give main thread plenty of time to exit
    }

    if (!called) {
	// The first time, don't destroy the window.  This allows Alt-W to
	// work without crashing the main thread.
	called = true;
	return false;
    }
    else {
	return true;
    }
}


void
GalagaWin::Zoom(BPoint rec_position, float rec_width, float rec_height)
{
    BPoint size(Bounds().Width(), Bounds().Height());
    if (_zoomed)
	size.x /= 2.0, size.y /= 2.0;
    else
	size.x *= 2.0, size.y *= 2.0;

    _zoomed = !_zoomed;

    BWindow::Zoom(Frame().LeftTop(), size.x, size.y);
    BView* view = ChildAt(0);
    view->SetScale(_zoomed ? 2.0 : 1.0);
}


Galaga::Galaga():
    BApplication("application/x-vnd.msteed-xgalaga"),
    _mainThread(NULL), _argc(0), _argv(NULL)
{
    extern char *app_dir;

    app_info info;
    GetAppInfo(&info);
    BEntry entry(&info.ref);
    entry.GetParent(&entry);

    BPath path;
    entry.GetPath(&path);
    app_dir = strdup(path.Path());
}


Galaga::~Galaga()
{
    extern char *app_dir;

    if (_argc > 0 && _argv != NULL) {
	for (int i = 0; i < _argc; i++) {
	    free(_argv[i]);
	}
	delete _argv;
    }

    _mainThread->requestQuit();

    free(app_dir);
}


void
Galaga::ArgvReceived(int32 argc, char **argv)
{
    _argc = argc;
    _argv = new char *[argc];
    for (int i = 0; i < argc; i++) {
	_argv[i] = strdup(argv[i]);
    }
}


void
Galaga::ReadyToRun()
{
    GalagaWin *w = new GalagaWin();
    _mainThread = new MainThread();
    BMessage msg(kMsgGo);
    msg.AddInt32("argc", _argc);
    msg.AddPointer("argv", _argv);
    _mainThread->PostMessage(&msg);
}


void
MainThread::requestQuit()
{
    Lock();
    Quit();
}


void
MainThread::MessageReceived(BMessage *m)
{
    switch (m->what) {
	case kMsgGo:
	{
	    int32 argc;
	    char **argv;
	    m->FindInt32("argc", &argc);
	    m->FindPointer("argv", (void**)&argv);
	    original_main(argc, argv);

	    be_app->PostMessage(B_QUIT_REQUESTED);
	    break;
	}
	default:
	    BLooper::MessageReceived(m);
	    break;
    }
}


int
main()
{	
    new Galaga();
    be_app->Run();
    delete be_app;

    return 0;
}

